#include "wrap/thread.h"
#include "wrap/synch.h"
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

/**
 * Efter att ha experimenterat med buffrar och insett att de inte passade ditt
 * problem särskilt bra blev du nyfiken på ifall det fanns andra lösningar som
 * passar bättre. Efter lite letande hittade du något som heter "future" som
 * verkar som att det kan lösa problemet. Naturligtvis börjar du implementera en
 * future för att se hur väl den kan lösa ditt problem!  Dock verkar inte
 * implementationen fungera som den ska.
 *
 * En future är en datatyp som representerar ett resultat som kommer att finnas
 * någon gång i framtiden. Den har följande operationer:
 *
 * "future_create" - skapar en future.
 *
 * "future_destroy" - förstör en future.
 *
 * "future_post" - meddelar future-objektet att det finns ett resultat
 * tillgängligt. Future-objektet lagrar resultatet till senare, och kommer ihåg
 * att det finns ett resultat tillgängligt. Du kan anta att "future_post" bara
 * anropas en gång för varje future-objekt.
 *
 * "future_get" - försöker hämta resultatet som är lagrat i
 * future-objektet. Ifall resultatet inte finns än ska den vänta på att en annan
 * tråd anropar "future_post". Finns resultatet ska den helt enkelt returnera
 * det resultat som sattes via "future_post". "future_get" ska kunna anropas
 * flera gånger på samma future-objekt.
 */

// Future-objektet.
struct future {
    // Resultatet som har lagrats här i.
    int result;
};

// Skapa ett future-objekt.
struct future *future_create(void) {
    // Allokera minne och initiera det.
    struct future *future = malloc(sizeof(struct future));
    future->result = 0;
    return future;
}

// Frigör minnet.
void future_destroy(struct future *future) {
    free(future);
}


// Meddelar future-objektet att det finns ett resultat
// tillgängligt. Future-objektet lagrar resultatet till senare, och kommer ihåg
// att det finns ett resultat tillgängligt. Du kan anta att "future_post" bara
// anropas en gång för varje future-objekt.
void future_post(struct future *future, int value) {
    future->result = value;
}

// Försöker hämta resultatet som är lagrat i future-objektet. Ifall resultatet
// inte finns än ska den vänta på att en annan tråd anropar "future_post". Finns
// resultatet ska den helt enkelt returnera det resultat som sattes via
// "future_post". "future_get" ska kunna anropas flera gånger på samma
// future-objekt.
int future_get(struct future *future) {
    return future->result;
}

/**
 * Huvudprogram som använder future-objektet. Har du gjort rätt ovan ska inga
 * ändringar behöva göras här, men det kan vara intressant göra ändringar för
 * att testa din implementation.
 */

// Funktion för beräkningar på en separat tråd.
void thread1(struct future *future) {
    // Heavy computations...
    future_post(future, 100);
}

// Funktion för beräkningar på en separat tråd.
void thread2(struct future *future) {
    // Heavy computations...
    future_post(future, 200);
}

int main(void) {
    // Skapa future-objekt.
    struct future *a = future_create();
    struct future *b = future_create();

    // Skapa trådar som sparar resultat i future-objekten.
    thread_new(&thread1, a);
    thread_new(&thread2, b);

    // Vänta på resultaten och skriv ut dem.
    printf("A: %d\n", future_get(a));
    printf("B: %d\n", future_get(b));
    printf("Sum: %d\n", future_get(a) + future_get(b));

    // Städa.
    future_destroy(a);
    future_destroy(b);

    return 0;
}
